import enum
import typing

from pg_sql import SqlId, SqlObject, SqlString, sql_list

from .formats.join import JoinTable
from .join_common import Structure
from .join_key import KeyResolver
from .string import indent


class _ChangeType(enum.Enum):
    CHANGE_1 = enum.auto()
    CHANGE_2 = enum.auto()


def create_change(
    id: str,
    table: JoinTable,
    table_id: str,
    resolver: KeyResolver,
    structure: Structure,
):
    change_1_function = structure.change_1_function(table_id)
    change_2_function = structure.change_2_function(table_id)

    for change_type in (_ChangeType.CHANGE_1, _ChangeType.CHANGE_2):
        if change_type == _ChangeType.CHANGE_1:
            change_function = change_1_function
        elif change_type == _ChangeType.CHANGE_2:
            change_function = change_2_function

        yield from _create_change_function(
            change_type=change_type,
            function=change_function,
            id=id,
            resolver=resolver,
            table=table,
            table_id=table_id,
        )

    delete_trigger = structure.delete_trigger(table_id)
    yield f"""
CREATE TRIGGER {delete_trigger} AFTER DELETE ON {table.sql}
REFERENCING OLD TABLE AS _change
FOR EACH STATEMENT EXECUTE PROCEDURE {change_1_function}()
    """.strip()

    insert_trigger = structure.insert_trigger(table_id)
    yield f"""
CREATE TRIGGER {insert_trigger} AFTER INSERT ON {table.sql}
REFERENCING NEW TABLE AS _change
FOR EACH STATEMENT EXECUTE PROCEDURE {change_1_function}()
    """.strip()

    update_trigger = structure.update_trigger(table_id)
    yield f"""
CREATE TRIGGER {update_trigger} AFTER UPDATE ON {table.sql}
REFERENCING OLD TABLE AS _old NEW TABLE AS _new
FOR EACH STATEMENT EXECUTE PROCEDURE {change_2_function}()
    """.strip()


def _create_change_function(
    change_type: _ChangeType,
    function: SqlObject,
    id: str,
    resolver: KeyResolver,
    table_id: str,
    table: JoinTable,
):
    def query(name: SqlObject):
        if table.columns is None:
            return f"TABLE {name}"
        values = sql_list(
            column.sql if column.value is None else f"{column.value} AS {column.sql}"
            for column in table.columns
        )
        return f"SELECT {values} FROM {name}"

    if change_type == _ChangeType.CHANGE_1:
        change = SqlObject("_change")
        root = f"({query(change)})"
    elif change_type == _ChangeType.CHANGE_2:
        old = SqlObject("_old")
        new = SqlObject("_new")
        root = f"""
(
    ({query(old)} EXCEPT ALL {query(new)})
    UNION ALL
    ({query(new)} EXCEPT ALL {query(old)})
)
    """.strip()

    yield f"""
CREATE FUNCTION {function} () RETURNS trigger
LANGUAGE plpgsql AS $$
  BEGIN
{indent(resolver.sql(root), 2)}

    RETURN NULL;
  END;
$$
    """.strip()

    yield f"""
COMMENT ON FUNCTION {function} IS {SqlString(f'Handle changes to {table_id} for {id}')}
    """.strip()



import copy
import json

from file import temp_file
from pg import connection, transaction
from process import run_process

_SCHEMA_SQL = """
    CREATE TABLE grandparent (
        id int PRIMARY KEY,
        name text NOT NULL
    );

    CREATE TABLE parent (
        id int PRIMARY KEY,
        grandparent_id int NOT NULL REFERENCES grandparent (id),
        name text NOT NULL
    );

    CREATE TABLE child (
        id int PRIMARY KEY,
        parent_id int NOT NULL REFERENCES parent (id)
    );

    CREATE TABLE child_full (
        id int PRIMARY KEY,
        grandparent_name text NOT NULL,
        parent_name text NOT NULL
    );
"""

_SCHEMA_JSON = {
    "context": ["context.example"],
    "id": "test",
    "tables": {
        "child": {
            "name": "child",
            "targetKey": ["child.id"],
        },
        "grandparent": {
            "join": "parent",
            "joinOn": "grandparent.id = parent.grandparent_id",
            "name": "grandparent",
        },
        "parent": {
            "key": [{"name": "id"}],
            "join": "child",
            "joinKey": ["id"],
            "joinMode": "async",
            "joinOn": "parent.id = child.parent_id",
            "name": "parent",
        },
    },
    "targetTable": {
        "key": ["id"],
        "columns": ["id", "grandparent_name", "parent_name"],
        "name": "child_full",
        "schema": "public",
    },
    "targetQuery": """
        SELECT c.id, g.name, p.name
        FROM ${key} AS d
            JOIN child c ON d.id = c.id
            JOIN parent p ON c.parent_id = p.id
            JOIN grandparent AS g ON p.grandparent_id = g.id
    """,
}


def test_join_async(pg_database):
    with temp_file("denorm-") as schema_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        output = run_process(
            [
                "denorm",
                "create-join",
                "--schema",
                schema_file,
            ]
        )
        with connection("") as conn, transaction(conn) as cur:
            # print(output.decode("utf-8"))
            cur.execute(output.decode("utf-8"))

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                    INSERT INTO grandparent (id, name)
                    VALUES (9, '_');

                    INSERT INTO parent (id, grandparent_id, name)
                    VALUES (1, 9, 'A'), (2, 9, 'B');

                    INSERT INTO child (id, parent_id)
                    VALUES (1, 1), (2, 1), (3, 2);
                """
            )

        with connection("") as conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                while True:
                    cur.execute("SELECT test__pcs__parent(10)")
                    (result,) = cur.fetchone()
                    if not result:
                        break

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM child_full ORDER BY id")
            result = cur.fetchall()
            assert result == [(1, "_", "A"), (2, "_", "A"), (3, "_", "B")]

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("UPDATE parent SET name = 'C' WHERE id = 2")

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM child_full ORDER BY id")
            result = cur.fetchall()
            assert result == [(1, "_", "A"), (2, "_", "A"), (3, "_", "B")]

        # import time
        # time.sleep(1000000)

        with connection("") as conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                while True:
                    cur.execute("SELECT test__pcs__parent(10)")
                    (result,) = cur.fetchone()
                    if not result:
                        break

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("TABLE test__que__parent")
            result = cur.fetchall()
            assert result == []

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM child_full ORDER BY id")
            result = cur.fetchall()
            assert result == [(1, "_", "A"), (2, "_", "A"), (3, "_", "C")]



from ..agg import AggIo, create_agg
from .common import open_str_read, open_str_write


def cli(args):
    io = AggIo(
        config=lambda: open_str_read(args.schema),
        output=lambda: open_str_write(args.output),
    )
    create_agg(io)



from ..join import JoinIo, create_join
from .common import open_str_read, open_str_write


def cli(args):
    io = JoinIo(
        config=lambda: open_str_read(args.schema),
        output=lambda: open_str_write(args.output),
    )
    create_join(io)



import typing

from pg_sql import SqlId, SqlNumber, SqlString, sql_list

from .agg_common import AggStructure
from .formats.agg import AggAggregate, AggConsistency, AggTable
from .string import indent


def create_change(
    aggregates: typing.Dict[str, AggAggregate],
    filter: typing.Optional[str],
    source: AggTable,
    id: str,
    consistency: AggConsistency,
    groups: typing.Dict[str, str],
    structure: AggStructure,
    target: AggTable,
):
    change_function = structure.change_function()
    group_columns = [SqlId(col) for col in groups]
    aggregate_columns = [SqlId(col) for col in aggregates]

    where = f"WHERE {filter}" if filter is not None else ""
    if consistency == AggConsistency.DEFERRED:
        setup_function = structure.setup_function()
        refresh_table = structure.refresh_table()
        target_table = structure.tmp_table()
        order = ""
        setup = f"""
PERFORM {setup_function}();
        """.strip()
        finalize = f"""
INSERT INTO {refresh_table}
SELECT
WHERE NOT EXISTS (TABLE {refresh_table});
        """.strip()
    elif consistency == AggConsistency.IMMEDIATE:
        target_table = target.sql
        order = f"ORDER BY {sql_list(SqlNumber(i + 1) for i, _ in enumerate(groups))}"
        setup = ""
        finalize = ""

    yield f"""
CREATE FUNCTION {change_function} () RETURNS trigger
LANGUAGE plpgsql AS $$
  DECLARE
    sign smallint := TG_ARGV[0]::smallint;
  BEGIN
{indent(setup, 2)}

    INSERT INTO {target_table} AS existing (
      {sql_list(group_columns)},
      {sql_list(aggregate_columns)}
    )
    SELECT
      {sql_list(value for value in groups.values())},
      {sql_list(agg.value for agg in aggregates.values())}
    FROM _change AS {SqlId(id)}
    {where}
    GROUP BY {sql_list(SqlNumber(i + 1) for i, _ in enumerate(groups))}
    {order}
    ON CONFLICT ({sql_list(group_columns)}) DO UPDATE
      SET {sql_list(f'{SqlId(col)} = {agg.combine_expression(col)}' for col, agg in aggregates.items())};

{indent(finalize, 2)}

    RETURN NULL;
  END;
$$
    """.strip()

    yield f"""
COMMENT ON FUNCTION {change_function} IS {SqlString(f'Handle changes for {id}')}
    """.strip()

    delete_trigger = structure.delete_trigger()
    yield f"""
CREATE TRIGGER {delete_trigger} AFTER DELETE ON {source.sql}
REFERENCING OLD TABLE AS _change
FOR EACH STATEMENT EXECUTE PROCEDURE {change_function}('-1');
    """.strip()

    insert_trigger = structure.insert_trigger()
    yield f"""
CREATE TRIGGER {insert_trigger} AFTER INSERT ON {source.sql}
REFERENCING NEW TABLE AS _change
FOR EACH STATEMENT EXECUTE PROCEDURE {change_function}('1');
    """.strip()

    update_1_trigger = structure.update_1_trigger()
    yield f"""
CREATE TRIGGER {update_1_trigger} AFTER UPDATE ON {source.sql}
REFERENCING NEW TABLE AS _change
FOR EACH STATEMENT EXECUTE PROCEDURE {change_function}('1');
    """.strip()

    update_2_trigger = structure.update_2_trigger()
    yield f"""
CREATE TRIGGER {update_2_trigger} AFTER UPDATE ON {source.sql}
REFERENCING OLD TABLE AS _change
FOR EACH STATEMENT EXECUTE PROCEDURE {change_function}('-1');
    """.strip()



import copy
import json

from file import temp_file
from pg import connection, transaction
from process import run_process

_SCHEMA_SQL = """
    CREATE TABLE child (
        id int PRIMARY KEY,
        parent_id int
    );

    CREATE TABLE parent_child_stat (
        parent_id int PRIMARY KEY,
        _count bigint NOT NULL,
        child_count int NOT NULL
    );
"""

_SCHEMA_JSON = {
    "id": "test",
    "source": {"name": "child"},
    "target": {"name": "parent_child_stat"},
    "groups": {"parent_id": "parent_id"},
    "aggregates": {
        "child_count": {
            "value": "sign * count(*)",
        }
    },
}


def test_agg(pg_database):
    with temp_file("denorm-") as schema_file, connection("") as conn:
        with transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        output = run_process(
            [
                "denorm",
                "create-agg",
                "--schema",
                schema_file,
            ]
        )
        with transaction(conn) as cur:
            cur.execute(output.decode("utf-8"))

        with transaction(conn) as cur:
            cur.execute(
                """
                    INSERT INTO child (id, parent_id)
                    VALUES (1, 1), (2, 1), (3, 2);
                """
            )

        with transaction(conn) as cur:
            cur.execute("SELECT * FROM parent_child_stat ORDER BY parent_id")
            result = cur.fetchall()
            assert result == [(1, 2, 2), (2, 1, 1)]


def test_agg_defer(pg_database):
    with temp_file("denorm-") as schema_file, connection("") as conn:
        with transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

        with open(schema_file, "w") as f:
            schema_json = copy.deepcopy(_SCHEMA_JSON)
            schema_json["consistency"] = "deferred"
            json.dump(schema_json, f)

        output = run_process(
            [
                "denorm",
                "create-agg",
                "--schema",
                schema_file,
            ]
        )
        with transaction(conn) as cur:
            cur.execute(output.decode("utf-8"))

        with transaction(conn) as cur:
            cur.execute(
                """
                    INSERT INTO child (id, parent_id)
                    VALUES (1, 1), (2, 1), (3, 2);
                """
            )
            cur.execute("TABLE parent_child_stat")
            result = cur.fetchall()
            assert result == []

        with transaction(conn) as cur:
            cur.execute("SELECT * FROM parent_child_stat ORDER BY parent_id")
            result = cur.fetchall()
            assert result == [(1, 2, 2), (2, 1, 1)]



import enum
import typing


class _State(enum.Enum):
    TEXT = enum.auto()
    REPLACEMENT_START = enum.auto()
    REPLACEMENT = enum.auto()


def format(format: str, replacements: typing.Dict[str, str]):
    i = 0
    state = _State.TEXT
    replacement = ""
    result = ""
    while True:
        if state == _State.TEXT:
            if len(format) <= i:
                break
            elif format[i] == "$":
                state = _State.REPLACEMENT_START
                i += 1
            else:
                result += format[i]
                i += 1
        elif state == _State.REPLACEMENT_START:
            if len(format) <= i:
                raise RuntimeError(f"Expected {{ at end of input")
            elif format[i] == "$":
                result += "$"
                state = _State.TEXT
                i += 1
            elif format[i] != "{":
                raise RuntimeError(f"Expected {{ at character {i}")
            else:
                state = _State.REPLACEMENT
                i += 1
        elif state == _State.REPLACEMENT:
            if len(format) <= i:
                raise RuntimeError(f"Expected }} at end of input")
            elif format[i] == "}":
                try:
                    result += replacements[replacement]
                except KeyError:
                    raise RuntimeError(f"Invalid param '{replacement}'")
                replacement = ""
                state = _State.TEXT
                i += 1
            else:
                replacement += format[i]
                i += 1

    return result



import typing

from pg_sql import SqlId, SqlNumber, SqlObject, SqlString, sql_list

from .format import format
from .join_common import JoinTarget, Key, Structure
from .join_key import KeyConsumer, TargetRefresh
from .sql import SqlTableExpr
from .sql_query import sync_query, upsert_query
from .string import indent


def create_refresh_function(
    id: str,
    structure: Structure,
    refresh: TargetRefresh,
):
    refresh_function = structure.refresh_function()
    refresh_table = structure.refresh_table()

    key_table = structure.key_table()
    refresh_sql = refresh.sql(f"TABLE {key_table}", None)

    yield f"""
CREATE FUNCTION {refresh_function} () RETURNS trigger
LANGUAGE plpgsql AS $$
  BEGIN
    -- analyze
    ANALYZE {refresh_table};

    -- refresh
{indent(str(refresh_sql), 2)}

    -- clear refresh
    DELETE FROM {refresh_table};

    RETURN NULL;
  END;
$$
""".strip()

    yield f"""
COMMENT ON FUNCTION {refresh_function} IS {SqlString(f'Refresh {id}')}
    """.strip()


def create_setup_function(
    structure: Structure,
    id: str,
    key: Key,
    target: JoinTarget,
):
    key_table = structure.key_table()
    refresh_constraint = structure.refresh_constraint()
    refresh_function = structure.refresh_function()
    refresh_table = structure.refresh_table()
    setup_function = structure.setup_function()

    yield f"""
CREATE FUNCTION {setup_function} () RETURNS void
LANGUAGE plpgsql AS $$
  BEGIN
    IF to_regclass({SqlString(str(refresh_table))}) IS NOT NULL THEN
      RETURN;
    END IF;

    CREATE TEMP TABLE {key_table}
    ON COMMIT DELETE ROWS
    AS {key.definition}
    WITH NO DATA;

    ALTER TABLE {key_table}
      ADD PRIMARY KEY ({sql_list([SqlId(name) for name in key.names])});

    CREATE TEMP TABLE {refresh_table} (
    ) ON COMMIT DELETE ROWS;

    CREATE CONSTRAINT TRIGGER {refresh_constraint} AFTER INSERT ON {refresh_table}
    DEFERRABLE INITIALLY DEFERRED
    FOR EACH ROW EXECUTE PROCEDURE {refresh_function}();
  END
$$
    """.strip()

    yield f"""
COMMENT ON FUNCTION {setup_function} IS {SqlString(f"Set up temp tables for {id}")}
    """.strip()


class DeferredKeys(KeyConsumer):
    def __init__(self, key: typing.List[str], structure: Structure):
        self._key = key
        self._structure = structure

    def sql(
        self,
        key_query: str,
        table_id: str,
        exprs: typing.List[SqlTableExpr] = [],
        last_expr: typing.Optional[str] = None,
    ):
        setup_function = self._structure.setup_function()

        refresh_table = self._structure.refresh_table()

        query = upsert_query(
            columns=self._key,
            key=self._key,
            query=key_query,
            target=self._structure.key_table(),
        )
        for expr in reversed(exprs):
            query.prepend(expr)
        if last_expr is not None:
            query.append(SqlId("_other"), last_expr)

        return f"""
PERFORM {setup_function}();

{query};

INSERT INTO {refresh_table}
SELECT
WHERE NOT EXISTS (TABLE {refresh_table});
        """.strip()



import dataclasses
import typing

from .agg_change import create_change
from .agg_clean import create_cleanup
from .agg_common import AggStructure
from .agg_defer import create_refresh_function, create_setup_function
from .formats.agg import AGG_DATA_JSON_FORMAT, AggAggregate, AggConfig, AggConsistency
from .resource import ResourceFactory


@dataclasses.dataclass
class AggIo:
    config: ResourceFactory[typing.TextIO]
    output: ResourceFactory[typing.TextIO]


def create_agg(io: AggIo):
    schema = AGG_DATA_JSON_FORMAT.load(io.config)

    with io.output() as f:
        for statement in _statements(schema):
            print(f"{statement};\n", file=f)


def _statements(config: AggConfig):
    structure = AggStructure(config.schema, config.id)

    config.aggregates["_count"] = AggAggregate(value="sign * count(*)")

    if config.consistency == AggConsistency.DEFERRED:
        yield from create_refresh_function(
            aggregates=config.aggregates,
            groups=config.groups,
            id=config.id,
            structure=structure,
            target=config.target,
        )

        yield from create_setup_function(
            aggregates=config.aggregates,
            groups=config.groups,
            id=config.id,
            structure=structure,
            target=config.target,
        )

    yield from create_change(
        aggregates=config.aggregates,
        consistency=config.consistency,
        filter=config.filter,
        groups=config.groups,
        id=config.id,
        source=config.source,
        structure=structure,
        target=config.target,
    )

    yield from create_cleanup(
        id=config.id, groups=config.groups, structure=structure, target=config.target
    )



import typing

from pg_sql import SqlObject, sql_list

from .format import format
from .join_common import JoinTarget, Key
from .sql import SqlQuery


class JoinPlainTarget(JoinTarget):
    """
    Insert into to a table
    """

    def __init__(self, query: str):
        self._query = query

    def key(self) -> typing.Optional[Key]:
        pass

    def sql(self, key_table: SqlObject, table_id: typing.Optional[str]) -> SqlQuery:
        formatted = format(
            self._query, {"key": str(key_table), "table": table_id or ""}
        )
        return SqlQuery(formatted)



import typing

from pg_sql import SqlId, SqlNumber, SqlString, sql_list

from .agg_common import AggStructure
from .formats.agg import AggAggregate, AggConfig, AggTable


def create_refresh_function(
    id: str,
    structure: AggStructure,
    aggregates: typing.Dict[str, AggAggregate],
    groups: typing.Dict[str, str],
    target: AggTable,
):
    refresh_function = structure.refresh_function()
    refresh_table = structure.refresh_table()
    tmp_table = structure.tmp_table()
    group_columns = [SqlId(col) for col in groups]
    aggregate_columns = [SqlId(col) for col in aggregates]

    yield f"""
CREATE FUNCTION {refresh_function} () RETURNS trigger
LANGUAGE plpgsql AS $$
  BEGIN
    DELETE FROM {refresh_table};

    WITH
      _delete AS (
        DELETE FROM {tmp_table}
        RETURNING *
      )
    INSERT INTO {target.sql} AS existing (
      {sql_list(group_columns)},
      {sql_list(aggregate_columns)}
    )
    SELECT
      {sql_list(group_columns)},
      {sql_list(aggregate_columns)}
    FROM {tmp_table}
    ORDER BY {sql_list(SqlNumber(i + 1) for i, _ in enumerate(groups))}
    ON CONFLICT ({sql_list(group_columns)}) DO UPDATE
      SET {sql_list(f'{SqlId(col)} = {agg.combine_expression(col)}' for col, agg in aggregates.items())};

    RETURN NULL;
  END;
$$
    """.strip()


def create_setup_function(
    id: str,
    structure: AggStructure,
    aggregates: typing.Dict[str, AggAggregate],
    groups: typing.Dict[str, str],
    target: AggTable,
):
    refresh_constraint = structure.refresh_constraint()
    refresh_function = structure.refresh_function()
    refresh_table = structure.refresh_table()
    setup_function = structure.setup_function()
    tmp_table = structure.tmp_table()

    group_columns = [SqlId(col) for col in groups]
    aggregate_columns = [SqlId(col) for col in aggregates]

    yield f"""
CREATE FUNCTION {setup_function} () RETURNS void
LANGUAGE plpgsql AS $$
  BEGIN
    IF to_regclass({SqlString(str(refresh_table))}) IS NOT NULL THEN
      RETURN;
    END IF;

    CREATE TEMP TABLE {tmp_table}
    ON COMMIT DELETE ROWS
    AS SELECT
      {sql_list(group_columns)},
      {sql_list(aggregate_columns)}
    FROM {target.sql}
    WITH NO DATA;

    ALTER TABLE {tmp_table}
      ADD PRIMARY KEY ({sql_list(group_columns)});

    CREATE TEMP TABLE {refresh_table} (
    ) ON COMMIT DELETE ROWS;

    CREATE CONSTRAINT TRIGGER {refresh_constraint} AFTER INSERT ON {refresh_table}
    DEFERRABLE INITIALLY DEFERRED
    FOR EACH ROW EXECUTE PROCEDURE {refresh_function}();
  END;
$$
    """.strip()

    yield f"""
COMMENT ON FUNCTION {setup_function} IS {SqlString(f"Set up temp tables for {id}")}
    """.strip()



import dataclasses
import enum
import typing

import dataclasses_json
from pg_sql import SqlId, SqlObject

from ..json import DataJsonFormat, ValidatingDataJsonFormat, package_json_format


class JoinConsistency(enum.Enum):
    DEFERRED = "deferred"
    IMMEDIATE = "immediate"


class JoinJoinMode(enum.Enum):
    ASYNC = "async"
    SYNC = "sync"


class JoinRefresh(enum.Enum):
    INSERT = "insert"
    FULL = "full"
    UPSERT = "upsert"


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
)
@dataclasses.dataclass
class JoinColumn:
    name: str
    value: typing.Optional[str] = None

    @property
    def sql(self) -> SqlId:
        return SqlId(self.name)


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
)
@dataclasses.dataclass
class JoinKeyColumn:
    name: str
    type: typing.Optional[str] = None
    value: typing.Optional[str] = None

    @property
    def sql(self) -> SqlId:
        return SqlId(self.name)


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
)
@dataclasses.dataclass
class JoinTargetTable:
    name: str
    columns: typing.Optional[typing.List[str]] = None
    key: typing.Optional[typing.List[str]] = None
    schema: typing.Optional[str] = None
    refresh: JoinRefresh = JoinRefresh.FULL

    @property
    def sql(self) -> SqlObject:
        return (
            SqlObject(SqlId(self.schema), SqlId(self.name))
            if self.schema is not None
            else SqlObject(SqlId(self.name))
        )


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
)
@dataclasses.dataclass
class JoinTable:
    columns: typing.Optional[typing.List[JoinColumn]] = None
    name: typing.Optional[str] = None
    join: typing.Optional[str] = None
    join_key: typing.Optional[typing.List[str]] = None
    join_on: typing.Optional[str] = None
    join_mode: JoinJoinMode = JoinJoinMode.SYNC
    join_other: typing.Optional[str] = None
    key: typing.Optional[typing.List[JoinKeyColumn]] = None
    key_type: typing.Optional[typing.List[str]] = None
    refresh_function: bool = False
    lock_id: typing.Optional[int] = None
    schema: typing.Optional[str] = None
    target_key: typing.Optional[typing.List[str]] = None

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(id(self))

    @property
    def sql(self) -> SqlObject:
        if self.name is None:
            return "(SELECT false AS _)"

        return (
            SqlObject(SqlId(self.schema), SqlId(self.name))
            if self.schema is not None
            else SqlObject(SqlId(self.name))
        )


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
)
@dataclasses.dataclass(frozen=True)
class JoinHook:
    name: str
    schema: typing.Optional[str] = None

    @property
    def sql(self) -> SqlObject:
        return (
            SqlObject(SqlId(self.schema), SqlId(self.name))
            if self.schema is not None
            else SqlObject(SqlId(self.name))
        )


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
)
@dataclasses.dataclass
class JoinConfig:
    id: str
    tables: typing.Dict[str, JoinTable]
    consistency: JoinConsistency = JoinConsistency.IMMEDIATE
    context: typing.List[str] = dataclasses.field(default_factory=list)
    key: typing.Optional[typing.List[JoinKeyColumn]] = None
    lock: bool = False
    schema: typing.Optional[str] = None
    setup: typing.Optional[JoinHook] = None
    target_query: typing.Optional[str] = "TABLE ${key}"
    target_table: typing.Optional[JoinTargetTable] = None


class JoinInvalid(Exception):
    def __init__(self, message):
        super().__init__(message)


def validate_join(join: JoinConfig):
    for table_id, table in join.tables.items():
        if table.refresh_function and any(
            column.type is None for column in table.key or []
        ):
            raise JoinInvalid(
                f"Table {table_id} refreshFunction required types for columns"
            )


JOIN_JSON_FORMAT = package_json_format("denorm.formats", "join.json")

JOIN_DATA_JSON_FORMAT = ValidatingDataJsonFormat(
    DataJsonFormat(JOIN_JSON_FORMAT, JoinConfig.schema()),
    validate_join,
)



import typing

from pg_sql import SqlObject, sql_list

from .format import format
from .formats.join import JoinRefresh, JoinTargetTable
from .join_common import JoinTarget, Key
from .sql import SqlQuery
from .sql_query import sync_query, upsert_query


class JoinTableTarget(JoinTarget):
    """
    Insert into to a table
    """

    def __init__(self, table: JoinTargetTable, query: str):
        self._table = table
        self._query = query

    def key(self) -> typing.Optional[Key]:
        if self._table.key:
            definition = f"""
SELECT {sql_list(self._table.key)}
FROM {self._table.sql}
            """.strip()
            names = self._table.key
            return Key(definition=definition, names=names)

    def sql(self, key_table: SqlObject, table_id: typing.Optional[str]) -> SqlQuery:
        formatted = format(
            self._query, {"key": str(key_table), "table": table_id or ""}
        )

        return sync_query(
            columns=self._table.columns or self._table.key,
            key=self._table.key,
            key_table=key_table,
            query=formatted,
            target=self._table.sql,
        )



import dataclasses
import typing

from pg_sql import SqlId, SqlObject

from .sql import SqlQuery


@dataclasses.dataclass
class Key:
    definition: str
    names: typing.List[str]


class Structure:
    def __init__(self, schema: str, id: str):
        self._schema = schema
        self._id = id

    def _name(self, name: str):
        return SqlId(f"{self._id}__{name}")

    def _sql_object(self, name: SqlId):
        return (
            SqlObject(SqlId(self._schema), name)
            if self._schema is not None
            else SqlObject(name)
        )

    def change_1_function(self, table_id: str):
        return self._sql_object(self._name(f"chg1__{table_id}"))

    def change_2_function(self, table_id: str):
        return self._sql_object(self._name(f"chg2__{table_id}"))

    def delete_trigger(self, table_id: str) -> SqlId:
        return self._name(f"del__{table_id}")

    def insert_trigger(self, table_id: str) -> SqlId:
        return self._name(f"ins__{table_id}")

    def update_trigger(self, table_id: str) -> SqlId:
        return self._name(f"upd__{table_id}")

    def lock_table(self) -> SqlObject:
        return self._sql_object(self._name("lock"))

    def queue_table(self, table_id: str) -> SqlObject:
        return self._sql_object(self._name(f"que__{table_id}"))

    def queue_process_function(self, table_id: str) -> SqlObject:
        return self._sql_object(self._name(f"pcs__{table_id}"))

    def key_table(self) -> SqlObject:
        return SqlObject(SqlId("pg_temp"), self._name("key"))

    def refresh_constraint(self) -> SqlId:
        return SqlId(self._id)

    def refresh_function(self) -> SqlObject:
        return self._sql_object(self._name("refresh"))

    def refresh_table(self) -> SqlObject:
        return SqlObject(SqlId("pg_temp"), self._name("refresh"))

    def refresh_table_function(self, table_id: str) -> SqlId:
        return self._sql_object(self._name(f"rfs__{table_id}"))

    def setup_function(self) -> SqlObject:
        return self._sql_object(self._name("setup"))


def context_column(column: str) -> str:
    return SqlId(f"context_{column}")


def local_column(column: str) -> str:
    return SqlId(f"local_{column}")


def foreign_column(column: str) -> str:
    return SqlId(f"foreign_{column}")


class JoinTarget(typing.Protocol):
    def key(self) -> typing.Optional[Key]:
        pass

    def sql(self, key_table: SqlObject, table_id: typing.Optional[str]) -> SqlQuery:
        pass



import typing

from pg_sql import SqlId, SqlObject


class AggStructure:
    def __init__(self, schema: str, id: str):
        self._schema = schema
        self._id = id

    def _name(self, name: str):
        return SqlId(f"{self._id}__{name}")

    def _sql_object(self, name: SqlId):
        return (
            SqlObject(SqlId(self._schema), name)
            if self._schema is not None
            else SqlObject(name)
        )

    def change_function(self) -> SqlObject:
        return self._sql_object(self._name("change"))

    def insert_trigger(self) -> SqlObject:
        return self._sql_object(self._name("ins"))

    def update_1_trigger(self) -> SqlObject:
        return self._sql_object(self._name("upd1"))

    def update_2_trigger(self) -> SqlObject:
        return self._sql_object(self._name("upd2"))

    def delete_trigger(self) -> SqlObject:
        return self._sql_object(self._name("del"))

    def cleanup_function(self) -> SqlObject:
        return self._sql_object(self._name("cleanup"))

    def cleanup_trigger(self) -> SqlId:
        return self._name("cleanup")

    def refresh_constraint(self) -> SqlId:
        return SqlId(self._id)

    def refresh_function(self) -> SqlObject:
        return self._sql_object(self._name("refresh"))

    def refresh_table(self) -> SqlObject:
        return SqlObject(SqlId("pg_temp"), self._name("refresh"))

    def setup_function(self) -> SqlObject:
        return self._sql_object(self._name("setup"))

    def tmp_table(self) -> SqlObject:
        return SqlObject(SqlId("pg_temp"), self._name("tmp"))



import copy
import json

from file import temp_file
from pg import connection, transaction
from process import run_process

_SCHEMA_SQL = """
    CREATE TABLE parent (
        id int PRIMARY KEY,
        name text NOT NULL
    );

    CREATE TABLE other (
        id int PRIMARY KEY,
        name text NOT NULL
    );

    CREATE TABLE other_override (
        id int PRIMARY KEY,
        parent_id int NOT NULL REFERENCES parent (id),
        other_id int NOT NULL REFERENCES other (id),
        name text NOT NULL
    );

    CREATE TABLE child (
        id int PRIMARY KEY,
        parent_id int NOT NULL REFERENCES parent (id)
    );

    CREATE TABLE grandchild (
        id int PRIMARY KEY,
        child_id int NOT NULL REFERENCES child (id),
        other_id int NOT NULL REFERENCES other (id)
    );

    CREATE TABLE child_full (
        id int PRIMARY KEY,
        parent_name text NOT NULL,
        other_names text[] NOT NULL
    );
"""

_SCHEMA_JSON = {
    "id": "test",
    "tables": {
        "child": {
            "name": "child",
            "targetKey": ["child.id"],
        },
        "grandchild": {
            "name": "grandchild",
            "targetKey": ["grandchild.child_id"],
        },
        "other": {
            "name": "other",
            "join": "grandchild",
            "joinOn": "other.id = grandchild.other_id",
        },
        "other_override": {
            "name": "other_override",
            "join": "grandchild",
            "joinOn": "(other_override.parent_id, other_override.other_id) = (child.parent_id, grandchild.other_id)",
            "joinOther": "JOIN child AS child ON child.id = grandchild.child_id",
        },
        "parent": {
            "join": "child",
            "joinOn": "parent.id = child.parent_id",
            "name": "parent",
        },
    },
    "targetTable": {
        "key": ["id"],
        "columns": ["id", "parent_name", "other_names"],
        "name": "child_full",
        "schema": "public",
    },
    "targetQuery": """
        SELECT c.id, p.name, gc.names
        FROM ${key} AS d
            JOIN child c ON d.id = c.id
            JOIN parent p ON c.parent_id = p.id
            CROSS JOIN LATERAL (
                SELECT coalesce(array_agg(coalesce(oo.name, o.name)), '{}') AS names
                FROM grandchild AS gc
                  JOIN other AS o ON gc.other_id = o.id
                  LEFT JOIN other_override AS oo ON (c.parent_id, gc.other_id) = (oo.parent_id, oo.other_id)
                WHERE c.id = gc.child_id
            ) AS gc
    """,
}


def test_join(pg_database):
    with temp_file("denorm-") as schema_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        output = run_process(
            [
                "denorm",
                "create-join",
                "--schema",
                schema_file,
            ]
        )
        # print(output.decode("utf-8"))
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(output.decode("utf-8"))

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                    INSERT INTO parent (id, name)
                    VALUES (1, 'A'), (2, 'B');

                    INSERT INTO other (id, name)
                    VALUES (1, 'Other');

                    INSERT INTO child (id, parent_id)
                    VALUES (1, 1), (2, 1), (3, 2);

                    INSERT INTO grandchild (id, child_id, other_id)
                    VALUES (1, 2, 1);
                """
            )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM child_full ORDER BY id")
            result = cur.fetchall()
            assert result == [(1, "A", []), (2, "A", ["Other"]), (3, "B", [])]

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                    INSERT INTO other_override (id, parent_id, other_id, name)
                    VALUES (1, 1, 1, 'Override');
                """
            )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM child_full ORDER BY id")
            result = cur.fetchall()
            assert result == [(1, "A", []), (2, "A", ["Override"]), (3, "B", [])]



"""
Procedures:
* ID__refresh - Perform refresh
  - When consistency is deferred
* ID__setup - Create the temporary tables
  - When consistency is deferred
* ID__chg1__SOURCE - Process changes
* ID__chg2__SOURCE - Process changes

Tables:
* BASE (existing) - Table to watch
  - Existing
  - Triggers
    * ID__del__SOURCE - Record deletes
    * ID__ins__SOURCE - Record inserts
    * ID__upd__SOURCE - Record updates
* TARGET (existing) - Table to populate
* ID__iterate__SOURCE - Queue changes for iteration
  - When iteration is used
* ID__lock - Value lock

Temp tables:
* ID__key - Keys to update
  - When consistency is deferred
* ID__refresh - Fire constraint trigger at end of transaction
  - When consistency is deferred
  - Triggers:
      * join (deferred) - Perform refresh
"""


import dataclasses
import typing

from pg_sql import SqlId, sql_list

from .format import format
from .formats.join import (
    JOIN_DATA_JSON_FORMAT,
    JoinConfig,
    JoinConsistency,
    JoinJoinMode,
)
from .join_async import create_queue
from .join_change import create_change
from .join_common import JoinTarget, Key, Structure
from .join_defer import DeferredKeys, create_refresh_function, create_setup_function
from .join_key import KeyResolver, TargetRefresh
from .join_lock import create_lock_table
from .join_plain_target import JoinPlainTarget
from .join_refresh_function import (
    create_refresh_function as create_table_refresh_function,
)
from .join_table_target import JoinTableTarget
from .resource import ResourceFactory
from .string import indent


@dataclasses.dataclass
class JoinIo:
    config: ResourceFactory[typing.TextIO]
    output: ResourceFactory[typing.TextIO]


def create_join(io: JoinIo):
    schema = JOIN_DATA_JSON_FORMAT.load(io.config)

    with io.output() as f:
        for statement in _statements(schema):
            print(f"{statement};\n", file=f)


def _target(config: JoinConfig) -> JoinTarget:
    if config.target_table:
        return JoinTableTarget(config.target_table, config.target_query)
    else:
        return JoinPlainTarget(config.target_query)


def _statements(config: JoinConfig):
    structure = Structure(config.schema, config.id)

    target = _target(config)
    key = target.key()
    if key is None:
        definition = f"SELECT {sql_list(f'NULL::{column.type} AS {column.sql}' for column in config.key)}"
        names = [column.name for column in config.key]
        key = Key(definition=definition, names=names)

    if config.lock:
        yield from create_lock_table(
            structure=structure, key=key, target=config.target_table
        )

    refresh_action = TargetRefresh(
        key=key.names,
        setup=config.setup,
        structure=structure,
        lock=config.lock,
        target=target,
    )

    if config.consistency == JoinConsistency.DEFERRED:
        yield from create_refresh_function(
            id=config.id,
            structure=structure,
            refresh=refresh_action,
        )

        yield from create_setup_function(
            structure=structure,
            id=config.id,
            target=config.target_table,
            key=key,
        )

    for table_id, table in config.tables.items():
        if table.join_mode != JoinJoinMode.ASYNC:
            continue

        resolver = KeyResolver(
            action=refresh_action,
            context=config.context,
            key=key.names,
            structure=structure,
            table_id=table.join,
            tables=config.tables,
        )

        yield from create_queue(
            context=config.context,
            id=config.id,
            resolver=resolver,
            structure=structure,
            table_id=table_id,
            tables=config.tables,
        )

    for table_id, table in config.tables.items():
        if config.consistency == JoinConsistency.DEFERRED:
            action = DeferredKeys(key=key.names, structure=structure)
        elif config.consistency == JoinConsistency.IMMEDIATE:
            action = refresh_action

        resolver = KeyResolver(
            action=action,
            context=config.context,
            key=key.names,
            structure=structure,
            table_id=table_id,
            tables=config.tables,
        )

        if table.refresh_function:
            yield from create_table_refresh_function(
                resolver=resolver,
                structure=structure,
                table=table,
                table_id=table_id,
            )

        if table.name is not None:
            yield from create_change(
                id=config.id,
                resolver=resolver,
                structure=structure,
                table=table,
                table_id=table_id,
            )



import typing

from pg_sql import SqlId, SqlNumber, SqlObject, SqlString, sql_list

from .format import format
from .formats.join import (
    JoinConsistency,
    JoinHook,
    JoinJoinMode,
    JoinRefresh,
    JoinTable,
    JoinTargetTable,
)
from .graph import closure
from .join_common import JoinTarget, Structure, foreign_column, local_column
from .join_lock import lock_sql
from .sql import SqlQuery, SqlTableExpr, table_fields, update_excluded
from .sql_query import sync_query, upsert_query
from .string import indent


class KeyConsumer(typing.Protocol):
    def sql(
        key_query: str,
        table_id: str,
        exprs: typing.List[SqlTableExpr] = [],
        last_expr: typing.Optional[SqlTableExpr] = None,
    ) -> str:
        pass


class TargetRefresh(KeyConsumer):
    def __init__(
        self,
        setup: typing.Optional[JoinHook],
        lock: bool,
        key: typing.List[str],
        structure: Structure,
        target: JoinTarget,
    ):
        self._key = key
        self._lock = lock
        self._setup = setup
        self._structure = structure
        self._target = target

    def sql(
        self,
        key_query: str,
        table_id: str,
        exprs: typing.List[SqlTableExpr] = [],
        last_expr: typing.Optional[str] = None,
    ):
        if self._setup is None:
            setup_sql = ""
        else:
            setup_sql = f"PERFORM {self._setup.sql}();"

        if self._lock:
            # Ideally, there would be a runtime-check and only use the lock table
            # if current_setting('transaction_isolation') == 'read committed'.
            # However, this isn't set up well to implement.
            lock_table = self._structure.lock_table()

            inner = f"""
ANALYZE {lock_table};

{target_query};
            """.strip()

            target_query = self._target.sql(lock_table, table_id)
            for expr in reversed(exprs):
                target_query.prepend(expr)
            if last_expr is not None:
                target_query.append(SqlId("other_"), last_expr)

            return f"""
{setup_sql}

{lock_sql(structure, self._key.names, inner)}
            """.strip()
        else:
            key_table = SqlId("_key")
            target_query = self._target.sql(key_table, table_id)
            target_query.prepend(SqlTableExpr(key_table, key_query))
            for expr in reversed(exprs):
                target_query.prepend(expr)
            if last_expr is not None:
                target_query.append(SqlId("other_"), last_expr)
            return f"""
{setup_sql}

{target_query};
            """.strip()


class KeyResolver:
    def __init__(
        self,
        action: KeyConsumer,
        key: typing.List[str],
        context: typing.List[str],
        structure: Structure,
        table_id: str,
        tables: typing.Dict[str, JoinTable],
    ):
        self._action = action
        self._context = context
        self._key = key
        self._structure = structure
        self._tables = tables
        self._table_id = table_id

        dep_ids = closure(
            [table_id],
            lambda id: [tables[id].join]
            if tables[id].join_mode == JoinJoinMode.SYNC and tables[id].join
            else [],
        )
        self._deps = [(id, tables[id]) for id in dep_ids]

    def sql(
        self,
        root: SqlObject,
        exprs: typing.List[SqlTableExpr] = [],
        last_expr: typing.Optional[str] = None,
    ) -> str:
        key_query = ""
        for i, (dep_id, dep) in enumerate(reversed(self._deps)):
            table_sql = root if i == len(self._deps) - 1 else str(dep.sql)

            if dep.target_key is not None:
                key_query += f"SELECT DISTINCT {sql_list(f'{k} AS {SqlId(t)}' for t, k in zip(self._key, dep.target_key))}"
                key_query += f"\nFROM"
                key_query += f"\n  {table_sql} AS {SqlId(dep_id)}"
            elif dep.join_mode == JoinJoinMode.ASYNC:
                if dep.key:
                    dep_columns = [column.sql for column in dep.key]
                    key_query += (
                        f"SELECT DISTINCT {table_fields(SqlId(dep_id), dep_columns)}"
                    )
                else:
                    key_query = f"SELECT {SqlId(dep_id)}.*"
                key_query += f"\nFROM"
                key_query += f"\n  {table_sql} AS {SqlId(dep_id)}"
            else:
                if dep.join_other is not None:
                    key_query += f"\n  {dep.join_other}"
                key_query += f"\n  JOIN {table_sql} AS {SqlId(dep_id)} ON {dep.join_on}"

        last_id, last_table = self._deps[-1]
        if last_table.join_mode == JoinJoinMode.ASYNC:
            from .join_async import enqueue_sql

            return enqueue_sql(
                context=self._context,
                id=last_id,
                table=last_table,
                structure=self._structure,
                key_query=key_query,
                exprs=exprs,
                last_expr=last_expr,
            )

        return self._action.sql(key_query, last_id, exprs=exprs, last_expr=last_expr)



import typing

T = typing.TypeVar("T")


class CycleError(Exception, typing.Generic[T]):
    def __init__(self, elements: typing.List[T]):
        self.elements = elements
        elements_str = " -> ".join(map(str, self.elements))
        super().__init__(f"Cycle: {elements_str}")


def closure(start: typing.List[T], fn: typing.Callable[[T], typing.List[T]]):
    visited: typing.Set[T] = set()
    chain: typing.List[T] = []

    def visit(item: T):
        if item in chain:
            raise CycleError(chain[chain.index(item) :])

        if item in visited:
            return

        yield item

        chain.append(item)
        visited.add(item)
        for dep in fn(item):
            yield from visit(dep)
        chain.pop()

    for item in start:
        yield from visit(item)



import typing

from pg_sql import SqlId, SqlNumber, SqlObject, sql_list

from .sql import SqlQuery, SqlTableExpr, table_fields, update_excluded
from .string import indent


def sync_query(
    columns: typing.List[SqlId],
    key: typing.List[SqlId],
    key_table: SqlObject,
    query: str,
    target: SqlObject,
) -> SqlQuery:
    """
    Insert, update, and delete
    """
    data_columns = [SqlId(column) for column in columns if column not in key]
    is_temp = target.names[0].name == "pg_temp"

    if not is_temp:
        query += f"\nORDER BY {sql_list(SqlNumber(i + 1) for i, _ in enumerate(key))}"

    if data_columns:
        upsert_query = f"""
INSERT INTO {target} ({sql_list(columns)})
{query}
ON CONFLICT ({sql_list(key)}) DO UPDATE
    SET {update_excluded(data_columns)}
RETURNING {sql_list(key)}
        """.strip()
    elif not is_temp:
        upsert_query = f"""
INSERT INTO {target} ({sql_list(columns)})
{query}
ON CONFLICT ({sql_list(key)}) DO UPDATE
    SET {update_excluded(key)}
    WHERE false
RETURNING {sql_list(key)}
        """.strip()
    else:
        upsert_query = f"""
INSERT INTO {target} ({sql_list(columns)})
{query}
ON CONFLICT ({sql_list(key)}) DO NOTHING
"""
    upsert_expression = SqlTableExpr(SqlId("_upsert"), upsert_query)

    delete_query = f"""
DELETE FROM {target} AS t
USING {key_table} AS k
  LEFT JOIN _upsert AS u ON ({table_fields(SqlId('k'), key)}) = ({table_fields(SqlId('u'), key)})
WHERE
  ({table_fields(SqlId('t'), key)}) = ({table_fields(SqlId('k'), key)})
  AND u.* IS NOT DISTINCT FROM NULL
    """.strip()

    return SqlQuery(delete_query, expressions=[upsert_expression])


def upsert_query(
    columns: typing.List[SqlId],
    key: typing.List[SqlId],
    query: str,
    target: SqlObject,
) -> SqlQuery:
    """
    Upsert data
    """
    data_columns = [SqlId(column) for column in columns if column not in key]
    is_temp = target.names[0].name == "pg_temp"

    if not is_temp:
        query += f"\nORDER BY {sql_list(SqlNumber(i + 1) for i, _ in enumerate(key))}"

    if data_columns:
        upsert_query = f"""
INSERT INTO {target} ({sql_list(columns)})
{query}
ON CONFLICT ({sql_list(key)}) DO UPDATE
    SET {update_excluded(data_columns)}
        """.strip()
    elif not is_temp:
        upsert_query = f"""
INSERT INTO {target} ({sql_list(columns)})
{query}
ON CONFLICT ({sql_list(key)}) DO UPDATE
    SET {update_excluded(key)}
    WHERE false
        """.strip()
    else:
        upsert_query = f"""
INSERT INTO {target} ({sql_list(columns)})
{query}
ON CONFLICT ({sql_list(key)}) DO NOTHING
"""

    return SqlQuery(upsert_query)



import dataclasses
import enum
import typing

import dataclasses_json
from pg_sql import SqlId, SqlObject

from ..json import DataJsonFormat, ValidatingDataJsonFormat, package_json_format


class AggConsistency(enum.Enum):
    DEFERRED = "deferred"
    IMMEDIATE = "immediate"


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
    undefined=dataclasses_json.Undefined.EXCLUDE,
)
@dataclasses.dataclass
class AggAggregate:
    value: str
    combine: typing.Optional[str] = None

    def combine_expression(self, name):
        return (
            self.combine
            if self.combine is not None
            else f"existing.{SqlId(name)} + excluded.{SqlId(name)}"
        )


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
    undefined=dataclasses_json.Undefined.EXCLUDE,
)
@dataclasses.dataclass
class AggTable:
    name: str
    schema: typing.Optional[str] = None

    @property
    def sql(self) -> SqlObject:
        return (
            SqlObject(SqlId(self.schema), SqlId(self.name))
            if self.schema is not None
            else SqlObject(SqlId(self.name))
        )


@dataclasses_json.dataclass_json(
    letter_case=dataclasses_json.LetterCase.CAMEL,
    undefined=dataclasses_json.Undefined.EXCLUDE,
)
@dataclasses.dataclass
class AggConfig:
    groups: typing.Dict[str, str]
    id: str
    aggregates: typing.Dict[str, AggAggregate]
    source: AggTable
    target: AggTable
    consistency: AggConsistency = AggConsistency.IMMEDIATE
    filter: typing.Optional[str] = None
    schema: typing.Optional[str] = None


AGG_JSON_FORMAT = package_json_format("denorm.formats", "agg.json")

AGG_DATA_JSON_FORMAT = DataJsonFormat(AGG_JSON_FORMAT, AggConfig.schema())



import json

from file import temp_file
from pg import connection, transaction
from process import run_process

_SCHEMA_SQL = """
    CREATE TABLE author (
        id int PRIMARY KEY,
        name text NOT NULL
    );

    CREATE TABLE book (
        id int PRIMARY KEY,
        title text NOT NULL
    );

    CREATE TABLE book_author (
        id int PRIMARY KEY,
        book_id int NOT NULL REFERENCES book (id) ON DELETE CASCADE,
        author_id int NOT NULL REFERENCES author (id) ON DELETE CASCADE,
        ordinal int NOT NULL,
        UNIQUE (book_id, author_id),
        UNIQUE (book_id, ordinal)
    );

    CREATE TABLE book_full (
        id int PRIMARY KEY,
        title text NOT NULL,
        author_names text[] NOT NULL
    );
"""

_SCHEMA_JSON = {
    "id": "book_full",
    "schema": "public",
    "tables": {
        "author": {
            "join": "book_author",
            "joinOn": "author.id = book_author.author_id",
            "name": "book_author",
            "schema": "public",
        },
        "book": {"name": "book", "schema": "public", "targetKey": ["book.id"]},
        "book_author": {
            "name": "book_author",
            "schema": "public",
            "targetKey": ["book_author.book_id"],
        },
    },
    "targetTable": {
        "columns": ["id", "title", "author_names"],
        "name": "book_full",
        "schema": "public",
        "key": ["id"],
    },
    "targetQuery": """
    SELECT
        b.id,
        b.title,
        a.names
    FROM
        ${key} AS k (id)
        JOIN book AS b ON k.id = b.id
        CROSS JOIN LATERAL (
            SELECT coalesce(array_agg(a.name ORDER BY ba.ordinal), '{}') AS names
            FROM
                author AS a
                JOIN book_author AS ba ON a.id = ba.author_id
            WHERE b.id = ba.book_id
        ) AS a
    """,
}


def test_join_example(pg_database):
    with temp_file("denorm-") as schema_file:
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(_SCHEMA_SQL)

        with open(schema_file, "w") as f:
            json.dump(_SCHEMA_JSON, f)

        output = run_process(
            [
                "denorm",
                "create-join",
                "--schema",
                schema_file,
            ]
        )
        with connection("") as conn, transaction(conn) as cur:
            cur.execute(output.decode("utf-8"))

        with connection("") as conn, transaction(conn) as cur:
            cur.execute(
                """
                    INSERT INTO author (id, name)
                    VALUES
                        (1, 'Neil Gaiman'),
                        (2, 'Terry Pratchett');

                    INSERT INTO book (id, title)
                    VALUES
                        (1, 'Good Omens'),
                        (2, 'The Color of Magic');

                    INSERT INTO book_author (id, book_id, author_id, ordinal)
                    VALUES
                        (1, 1, 1, 1),
                        (2, 1, 2, 2),
                        (3, 2, 2, 1);
                """
            )

        with connection("") as conn, transaction(conn) as cur:
            cur.execute("SELECT * FROM book_full ORDER BY id")
            result = cur.fetchall()
            assert result == [
                (1, "Good Omens", ["Neil Gaiman", "Terry Pratchett"]),
                (2, "The Color of Magic", ["Terry Pratchett"]),
            ]



import hashlib
import typing

from pg_sql import SqlId, SqlNumber, SqlObject, SqlString, sql_list

from .formats.join import JoinTable
from .join_common import Structure, context_column, foreign_column, local_column
from .join_key import KeyResolver
from .sql import SqlQuery, SqlTableExpr, table_fields, update_excluded
from .string import indent


def create_queue(
    id: str,
    table_id: str,
    structure: Structure,
    resolver: KeyResolver,
    tables: typing.Dict[str, JoinTable],
    context: typing.List[str],
):
    table = tables[table_id]
    dep = table.join
    foreign_table = tables[dep]

    if table.lock_id is not None:
        lock_id = table.lock_id
    else:
        digest = hashlib.md5(f"{id}__{table_id}".encode("utf-8")).digest()
        lock_id = int.from_bytes(digest[0:2], "big", signed=True)
    lock_base = lock_id * (2 ** 48)

    queue_table = structure.queue_table(table_id)

    column_names = [column.name for column in table.key] if table.key else ["_"]

    local_columns = [local_column(column) for column in column_names]
    foreign_columns = [foreign_column(column) for column in table.join_key]
    context_columns = [context_column(setting) for setting in context]

    columns = (
        [
            f"{SqlObject(SqlId('l'), SqlId(column))} AS {local_column(column)}"
            for column in column_names
        ]
        + [
            f"{SqlObject(SqlId('f'), SqlId(column))} AS {foreign_column(column)}"
            for column in table.join_key
        ]
        + [f"NULL::text AS {context_column(setting)}" for setting in context]
        + ["NULL::bigint AS seq", "NULL::bigint AS lock", "NULL::bigint AS count"]
    )

    yield f"""
CREATE TABLE {queue_table}
AS SELECT {sql_list(columns)}
FROM
  {table.sql} AS l
  CROSS JOIN {foreign_table.sql} AS f
WITH NO DATA
    """.strip()

    yield f"""
ALTER TABLE {queue_table}
  ADD PRIMARY KEY ({sql_list(local_columns + context_columns)}),
  ALTER count SET NOT NULL,
  ALTER count SET DEFAULT 0,
  ALTER lock ADD GENERATED BY DEFAULT AS IDENTITY,
  ALTER lock SET NOT NULL,
  ALTER seq ADD GENERATED BY DEFAULT AS IDENTITY,
  ALTER seq SET NOT NULL
    """.strip()

    yield f"""
COMMENT ON TABLE {queue_table} IS {SqlString(f"Asynchronous processing of changes to {table.sql}")}
    """.strip()

    for column in column_names:
        yield f"""
COMMENT ON COLUMN {queue_table}.{local_column(column)} IS {SqlString(f"{table.sql} key: {SqlId(column)}")}
"""

    for column in table.join_key:
        yield f"""
COMMENT ON COLUMN {queue_table}.{foreign_column(column)} IS {SqlString(f"{foreign_table.sql} iterator: {SqlId(column)}")}
"""

    yield f"""
COMMENT ON COLUMN {queue_table}.seq IS 'Order to process'
    """.strip()

    yield f"""
COMMENT ON COLUMN {queue_table}.lock IS 'Lock ID (add to base value {lock_base})'
    """.strip()

    yield f"""
COMMENT ON COLUMN {queue_table}.count IS 'Count of records processed'
    """.strip()

    yield f"""
CREATE INDEX ON {queue_table} (seq)
    """.strip()

    foreign_key_table = SqlObject(SqlId("_foreign_key"))

    item = SqlId("_item")
    new_item = SqlId("_new_item")

    new_fields = ["''" for _ in context] + ["0", "0", "_item.count + count(*) OVER ()"]
    get_item = f"""
SELECT
  {table_fields(item, local_columns)},
  {table_fields(SqlId("k"), [SqlId(column) for column in table.join_key])},
  {sql_list(new_fields)}
INTO _new_item
FROM {SqlObject(foreign_key_table)} AS k
ORDER BY {table_fields(SqlId("k"), table.join_key)} DESC
    """.strip()

    if table.join_on is not None:
        join = f"""
JOIN (VALUES ({table_fields(item, local_columns)})) AS {SqlId(table_id)} ({sql_list(SqlId(col.name) for col in table.key)})
ON {table.join_on}
        """.strip()
    else:
        join = ""

    key1_query = f"""
SELECT {SqlId(dep)}.*
FROM {foreign_table.sql} AS {SqlId(dep)}
{join}
ORDER BY {sql_list(SqlObject(SqlId(dep), SqlId(name)) for name in table.join_key)}
LIMIT max_records
    """.strip()
    gather1 = resolver.sql(
        foreign_key_table,
        exprs=[SqlTableExpr(foreign_key_table, key1_query)],
        last_expr=get_item,
    )

    key2_query = f"""
SELECT {SqlId(dep)}.*
FROM {foreign_table.sql} AS {SqlId(dep)}
{join}
WHERE ({table_fields(item, foreign_columns)}) < ({table_fields(SqlId(dep), (SqlId(column) for column in table.join_key))})
ORDER BY {sql_list(SqlObject(SqlId(dep), SqlId(name)) for name in table.join_key)}
LIMIT max_records
    """.strip()
    gather2 = resolver.sql(
        foreign_key_table,
        exprs=[SqlTableExpr(foreign_key_table, key2_query)],
        last_expr=get_item,
    )

    context_vars = "\n".join(
        f"{SqlId(f'_context_{setting}')} text := current_setting({SqlString(setting)}, true);"
        for setting in context
    )
    set_context = "\n".join(
        f"PERFORM set_config({SqlString(setting)}, _item.{context_column(setting)}, true);"
        for setting in context
    )
    unset_context = "\n".join(
        f"PERFORM set_config({SqlString(setting)}, {SqlId(f'_context_{setting}')}, true);"
        for setting in context
    )

    process_function = structure.queue_process_function(table_id)
    yield f"""
CREATE FUNCTION {process_function} (max_records bigint) RETURNS bool
LANGUAGE plpgsql AS $$
  DECLARE
    _item {queue_table};
    _new_item {queue_table};
    {context_vars}
  BEGIN
    -- find item
    SELECT (q.*) INTO _item
    FROM {queue_table} AS q
    WHERE pg_try_advisory_xact_lock({lock_base} + q.lock)
    ORDER BY q.seq
    LIMIT 1;

    IF _item IS NULL THEN
      -- if no item found, exit
      RETURN false;
    END IF;

{indent(set_context, 2)}

    IF ({table_fields(item, (foreign_column(column) for column in table.join_key))}) IS NULL THEN
      -- if there is no iterator, start at the beginning
{indent(gather1, 3)}
    ELSE
      -- if there is an iterator, start at the iterator
{indent(gather2, 3)}
    END IF;

    IF _new_item IS NULL THEN
      -- if the iterator was at the end, remove the queue item
      DELETE FROM {queue_table} AS q
      WHERE
        ({table_fields(SqlId("q"), local_columns + context_columns)}, q.seq)
          = ({table_fields(item, local_columns + context_columns)}, _item.seq);
    ELSE
      -- update the queue item with the new iterator
      UPDATE {queue_table} AS q
      SET
        {sql_list(f'{column} = (_new_item).{column}' for column in foreign_columns)},
        count = _new_item.count,
        seq = nextval(pg_get_serial_sequence({SqlString(str(queue_table))}, 'seq'))
      WHERE
          ({table_fields(SqlId("q"), local_columns)}, q.seq)
          = ({table_fields(item, local_columns)}, _item.seq);
    END IF;

{indent(unset_context, 2)}

    -- notify listeners that the queue has been updated
    NOTIFY {SqlId(str(queue_table))};

    RETURN true;
  END;
$$
""".strip()

    yield f"""
COMMENT ON FUNCTION {process_function} IS {SqlString(f"Refresh for {queue_table}")}
    """.strip()


def enqueue_sql(
    id: str,
    context: typing.List[str],
    table: JoinTable,
    structure: Structure,
    key_query: str,
    exprs: typing.List[SqlTableExpr],
    last_expr: typing.Optional[str],
):
    queue_table = structure.queue_table(id)

    column_names = [column.name for column in table.key] if table.key else ["_"]
    local_columns = [local_column(column) for column in column_names]
    context_columns = [context_column(setting) for setting in context]

    if table.key:
        order = (
            f"ORDER BY {sql_list(SqlNumber(i + 1) for i, _ in enumerate(table.key))}"
        )
    else:
        order = ""

    if context:
        settings = sql_list(
            f"coalesce(current_setting({SqlString(f'context.{setting}')}, true), '')"
            for setting in context
        )
        key_query = f"""
SELECT *, {settings}
FROM (
{indent(key_query, 1)}
) AS t
        """.strip()

    insert = f"""
INSERT INTO {queue_table} ({sql_list(local_columns + context_columns)})
{key_query}
{order}
ON CONFLICT ({sql_list(local_columns + context_columns)}) DO UPDATE
  SET {update_excluded(foreign_column(column) for column in table.join_key)},
    count = excluded.count,
    seq = excluded.seq
    """.strip()
    query = SqlQuery(insert, expressions=exprs)
    if last_expr is not None:
        query.append(SqlId("_other"), last_expr)

    return f"""
{query};

NOTIFY {SqlId(str(queue_table))};
    """.strip()



import typing

from pg_sql import SqlId, SqlString, sql_list

from .join_common import JoinTarget, Key, Structure
from .sql_query import upsert_query


def create_lock_table(structure: Structure, key: Key, target: JoinTarget):
    lock_table = structure.lock_table()

    yield f"""
CREATE UNLOGGED TABLE {lock_table}
AS {key.definition}
WITH NO DATA
    """.strip()

    yield f"""
ALTER TABLE {lock_table}
  ADD PRIMARY KEY ({sql_list(SqlId(name) for name in key.names)})
    """.strip()

    yield f"""
COMMENT ON TABLE {lock_table} IS {SqlString(f"Value lock on {target.name} primary key")}
    """.strip()


def lock_sql(structure: Structure, key: typing.List[str], sql: str):
    lock_table = structure.lock_table()
    lock_query = upsert_query(
        columns=key,
        key=key,
        query=key_query,
        target=lock_table,
    )

    return f"""
{lock_query};

{sql}

DELETE FROM {lock_table}
    """.strip()



from pg_sql import SqlId, SqlObject, SqlString

from .formats.join import JoinTable
from .join_async import enqueue_sql
from .join_common import Structure
from .join_key import KeyResolver
from .sql import SqlTableExpr, sql_list
from .string import indent


def param_name(name) -> SqlId:
    return SqlId(f"_{name}")


def create_refresh_function(
    structure: Structure, resolver: KeyResolver, table_id: str, table: JoinTable
):
    if table.key:
        columns = sql_list(
            f"{param_name(column.name)} AS {column.sql}" for column in table.key
        )
        key_query = f"SELECT {columns}"
    else:
        key_query = "SELECT false AS _"

    key = SqlId("key")
    query = resolver.sql(SqlObject(key), [SqlTableExpr(name=key, query=key_query)])

    function = structure.refresh_table_function(table_id)

    params = sql_list(
        f"{param_name(column.name)} {column.type}" for column in (table.key or [])
    )
    yield f"""
CREATE FUNCTION {function}({params}) RETURNS void
LANGUAGE plpgsql AS $$
  BEGIN
{indent(query, 2)}
  END;
$$
    """.strip()

    comment = SqlString(f"Recalculate, based on {table.join}")
    yield f"""
COMMENT ON FUNCTION {function} IS {comment}
    """.strip()



#!/usr/bin/env python3
import os

import setuptools

version = {}
with open("denorm/version.py", "r") as f:
    exec(f.read(), version)

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    author="Rivet Health",
    author_email="ops@rivethealth.com",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
    ],
    description="Maintain denormalized and aggregated PostgreSQL tables",
    entry_points={
        "console_scripts": [
            "denorm=denorm.cli.main:main",
        ]
    },
    extras_require={"dev": ["black", "jsonschema2md", "pytest-env", "isort", "pytest"]},
    package_data={
        "denorm.formats": ["*.json"],
    },
    install_requires=["dataclasses-json", "jsonschema", "pg-sql>=0.0.1"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="denorm",
    packages=setuptools.find_packages(),
    project_urls={
        "Issues": "https://github.com/rivethealth/denorm/issues",
    },
    url="https://github.com/rivethealth/denorm",
    version=version["__version__"],
)



import argparse

from ..version import __version__


def main():
    parser = _create_parser()
    args = parser.parse_args()

    if args.command == "create-agg":
        from .create_agg import cli

        cli(args)
    if args.command == "create-join":
        from .create_join import cli

        cli(args)


def _create_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="show version and exit",
    )

    subparsers = parser.add_subparsers(dest="command")

    _add_create_agg_command(subparsers)
    _add_create_join_command(subparsers)

    return parser


def _add_create_agg_command(subparsers):
    parser = subparsers.add_parser("create-agg")
    parser.add_argument("--schema", default="-")
    parser.add_argument("--output", default="-")


def _add_create_join_command(subparsers):
    parser = subparsers.add_parser("create-join")
    parser.add_argument("--schema", default="-")
    parser.add_argument("--output", default="-")



import typing

from pg_sql import SqlId, SqlString

from .agg_common import AggStructure
from .formats.agg import AggTable
from .sql import table_fields


def create_cleanup(
    id: str, groups: typing.Dict[str, str], structure: AggStructure, target: AggTable
):
    cleanup_function = structure.cleanup_function()
    cleanup_trigger = structure.cleanup_trigger()
    group_columns = [SqlId(group) for group in groups]

    yield f"""
CREATE FUNCTION {cleanup_function} () RETURNS trigger
LANGUAGE plpgsql AS $$
  BEGIN
    DELETE FROM {target.sql} AS t
    USING _new AS n
    WHERE
      ({table_fields(SqlId("t"), group_columns)}) = ({table_fields(SqlId("n"), group_columns)})
      AND n._count = 0;

    RETURN NULL;
  END;
$$
    """.strip()

    yield f"""
COMMENT ON FUNCTION {cleanup_function} IS {SqlString(f'Cleanup records for {id}')}
    """.strip()

    yield f"""
CREATE TRIGGER {cleanup_trigger} AFTER UPDATE ON {target.sql}
REFERENCING NEW TABLE AS _new
FOR EACH STATEMENT EXECUTE PROCEDURE {cleanup_function}()
    """.strip()

    yield f"""
COMMENT ON TRIGGER {cleanup_trigger} ON {target.sql} IS {SqlString(f'Cleanup records for {id}')}
    """.strip()



import collections
import dataclasses
import re
import typing

from pg_sql import SqlId, SqlObject, sql_list

from .string import indent


@dataclasses.dataclass
class SqlTableExpr:
    name: SqlId
    query: str

    def __str__(self) -> str:
        return f"""
{self.name} AS (
{indent(self.query, 1)}
)""".strip()


@dataclasses.dataclass
class SqlQuery:
    query: str
    expressions: typing.List[SqlTableExpr] = dataclasses.field(default_factory=list)

    def append(self, id: SqlId, query: str):
        self.expressions.append(SqlTableExpr(id, self.query))
        self.query = query

    def prepend(self, expression: SqlTableExpr):
        self.expressions.insert(0, expression)

    def __str__(self) -> str:
        result = ""
        if self.expressions:
            expressions = ",\n".join(str(expr) for expr in self.expressions)
            result += f"""
WITH
{indent(expressions, 1)}
        """.strip()
        result += "\n"
        return f"{result}{self.query}"


def update_excluded(columns: typing.Iterable[SqlId]):
    return sql_list(f"{column} = excluded.{column}" for column in columns)


def table_fields(id: SqlId, columns: typing.List[SqlId]):
    return sql_list(str(SqlObject(id, column)) for column in columns)
